"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.TypesGenerator = exports.SUPPORTED_REQUEST_CONTENT = void 0;
const openapi_schema_to_json_schema_1 = __importDefault(require("@openapi-contrib/openapi-schema-to-json-schema"));
const fs_1 = require("fs");
const promises_1 = require("fs/promises");
const json_schema_to_typescript_1 = require("json-schema-to-typescript");
const ts_morph_1 = require("ts-morph");
const ts_morph_2 = require("ts-morph");
const helpers_1 = require("./helpers");
exports.SUPPORTED_REQUEST_CONTENT = ['multipart/form-data', 'application/json'];
const isEmptyObject = (val) => {
    if (!val)
        return true;
    if (typeof val !== 'object')
        return false;
    return Object.keys(val).length === 0;
};
function convertToJsonSchema(schema) {
    const res = (0, openapi_schema_to_json_schema_1.default)(schema, {
        cloneSchema: false,
        strictMode: false,
        keepNotSupported: ['nullable'],
        beforeTransform: schema => {
            if (!('type' in schema))
                return schema;
            if (schema.type === 'int') {
                schema.type = 'integer';
            }
            if (Array.isArray(schema.type)) {
                if (schema.type.includes('null'))
                    return {
                        type: schema.type.filter(e => e !== 'null')[0],
                        nullable: true,
                    };
                else
                    throw new Error('Invalid type tuple: ' + JSON.stringify(schema));
            }
            return schema;
        },
    });
    delete res['$schema'];
    return res;
}
function extractResponseJsonSchema(schema) {
    if (schema.type !== 'object' || !schema.properties)
        throw new Error('Invalid json schema for response:' + JSON.stringify(schema));
    if (!('data' in schema.properties) && !('meta' in schema.properties))
        throw new Error('Invalid json schema for response:' + JSON.stringify(schema));
    let data = undefined;
    let meta = undefined;
    if ('data' in schema.properties) {
        data = schema.properties.data;
    }
    if ('meta' in schema.properties) {
        meta = schema.properties.meta;
    }
    return { data, meta };
}
class TypesGenerator {
    overridePolicy;
    constructor({ overridePolicy }) {
        this.overridePolicy = overridePolicy;
    }
    async generateRequestType(group, operation, imports) {
        const types = (0, helpers_1.extractOperationTypeNames)(operation);
        const isWithParam = (0, helpers_1.hasPathParams)(operation);
        if (!isEmptyObject(operation.requestBody?.content)) {
            const requestBody = operation.requestBody;
            const content = requestBody.content;
            if (content['application/json'] && content['multipart/form-data'])
                throw new Error('Invalid operation has both json and multipart content in request body: ' +
                    JSON.stringify(operation));
            if (content['application/json']?.schema && types.request) {
                const jsonSchema = convertToJsonSchema(content['application/json']?.schema);
                const ts = await (0, json_schema_to_typescript_1.compile)(jsonSchema, types.request, {
                    bannerComment: '',
                    additionalProperties: false,
                });
                return ts;
            }
            else if (content['multipart/form-data']?.schema) {
                return `export type ${types.request} = {
                    formData: FormData;
                    ${isWithParam ? 'id: number | string;' : ''}
                }`;
            }
            // toJsonSchema(operation.)
        }
        else if (isWithParam && types.request) {
            return `export type ${types.request} = { id: number | string };`;
        }
        return '';
    }
    async generateResponseType(group, operation, imports) {
        if (isEmptyObject(operation.responses))
            throw new Error('Invalid operation: ' + JSON.stringify(operation) + '. It has no responses.');
        const types = (0, helpers_1.extractOperationTypeNames)(operation);
        const responseDataSchema = {
            oneOf: [],
        };
        const responseMetaSchema = {
            oneOf: [],
        };
        Object.values(operation.responses).forEach(response => {
            if (response?.content?.['application/json']) {
                const jsonSchema = convertToJsonSchema(response?.content?.['application/json'].schema);
                try {
                    const { data, meta } = extractResponseJsonSchema(jsonSchema);
                    if (data)
                        responseDataSchema.oneOf.push(data);
                    if (meta)
                        responseMetaSchema.oneOf.push(meta);
                }
                catch (err) {
                    console.error('Invalid data at ', operation['x-path'], err);
                }
            }
            // TODO:
            // if (response?.content?.['application/octet-stream']) {
            //     const jsonSchema = convertToJsonSchema(response?.content?.['application/octet-stream']?.schema);
            //     oneOf.push(jsonSchema);
            // }
        });
        const dataCode = await (0, json_schema_to_typescript_1.compile)(responseDataSchema, types.responseData, {
            bannerComment: '',
            additionalProperties: false,
        });
        const metaCode = await (0, json_schema_to_typescript_1.compile)(responseMetaSchema, types.responseMeta, {
            bannerComment: '',
            additionalProperties: false,
        });
        return `
        ${dataCode}
        ${metaCode}
        export type ${types.response} = CommonResponse<${types.responseData}, ${types.responseMeta}>;
        `;
    }
    async generate(group, flatOperations) {
        const folder = `output/${group}`;
        await (0, promises_1.mkdir)(folder, { recursive: true });
        const filePath = `${folder}/types.ts`;
        const isExisting = (0, fs_1.existsSync)(filePath);
        if (this.overridePolicy === 'skip' && isExisting) {
            console.warn('Skipping', group, 'according to policy.');
            return;
        }
        const project = new ts_morph_1.Project();
        const sourceFile = project.createSourceFile(`types.ts`, '', { overwrite: true });
        const imports = [
            {
                from: '@api/common/types',
                name: 'CommonResponse',
            },
        ];
        const contentWriter = new ts_morph_2.CodeBlockWriter();
        await Promise.all(flatOperations.map(async (operation) => {
            const request = await this.generateRequestType(group, operation, imports);
            const response = await this.generateResponseType(group, operation, imports);
            contentWriter.write(request);
            contentWriter.blankLine();
            contentWriter.write(response);
        }));
        (0, helpers_1.renderImports)(sourceFile, imports);
        sourceFile.formatText();
        const content = sourceFile.getFullText() + '\n' + contentWriter.toString();
        await (0, promises_1.writeFile)(filePath, content);
    }
}
exports.TypesGenerator = TypesGenerator;
