"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.extractOperationTypeNames = exports.hasFileUpload = exports.hasPathParams = exports.renderImports = exports.groupOperations = exports.removeLeadingSlash = exports.flattenPaths = exports.isOperationMutation = exports.parseOpcode = exports.SEARCH_OPCODES = void 0;
const case_1 = require("case");
var HttpMethod;
(function (HttpMethod) {
    HttpMethod["GET"] = "get";
    HttpMethod["PUT"] = "put";
    HttpMethod["POST"] = "post";
    HttpMethod["DELETE"] = "delete";
    HttpMethod["OPTIONS"] = "options";
    HttpMethod["HEAD"] = "head";
    HttpMethod["PATCH"] = "patch";
    HttpMethod["TRACE"] = "trace";
})(HttpMethod || (HttpMethod = {}));
exports.SEARCH_OPCODES = ['search', 'get'];
const parseOpcode = (operation) => {
    const operationIdPath = (0, case_1.kebab)(operation.operationId).split('-');
    return operationIdPath[0];
};
exports.parseOpcode = parseOpcode;
const isOperationMutation = (method, operation) => {
    // assuming no mutations for these methods
    if ([HttpMethod.GET, HttpMethod.HEAD, HttpMethod.OPTIONS, HttpMethod.TRACE].includes(method))
        return false;
    const mainOperationCode = (0, exports.parseOpcode)(operation);
    return !exports.SEARCH_OPCODES.includes(mainOperationCode);
};
exports.isOperationMutation = isOperationMutation;
const flattenPaths = (paths) => {
    const pathNames = Object.keys(paths);
    return pathNames.flatMap(pathName => {
        const e = paths[pathName];
        const httpMethods = Object.keys(e);
        return httpMethods.map(httpMethod => ({
            ...e[httpMethod],
            'x-path': pathName,
            'x-method': httpMethod,
            isMutation: (0, exports.isOperationMutation)(httpMethod, e[httpMethod]),
        }));
    });
};
exports.flattenPaths = flattenPaths;
const removeLeadingSlash = (path) => path.replace('/', '');
exports.removeLeadingSlash = removeLeadingSlash;
const extractSegment = (path) => {
    const segments = path.split('/');
    if (segments.length < 2) {
        return undefined;
    }
    return segments[1];
};
const groupOperations = (flatOperation) => flatOperation.reduce((acc, cur) => {
    const groupName = extractSegment(cur['x-path']);
    if (!groupName)
        return acc;
    if (!(groupName in acc)) {
        acc[groupName] = [];
    }
    acc[groupName].push(cur);
    return acc;
}, {});
exports.groupOperations = groupOperations;
const renderImports = (sourceFile, imports) => {
    const map = new Map();
    imports.forEach(el => {
        if (!map.has(el.from))
            map.set(el.from, []);
        if (!map.get(el.from).find(e => e.name === el.name))
            map.get(el.from).push(el);
    });
    const importFroms = [...map.keys()];
    importFroms.forEach(importFrom => {
        const els = map.get(importFrom);
        if (els.length === 1 && els[0].isDefault) {
            sourceFile.addImportDeclaration({
                namespaceImport: els[0].name,
                moduleSpecifier: els[0].from,
            });
        }
        else {
            sourceFile.addImportDeclaration({
                namedImports: els.map(e => e.name),
                moduleSpecifier: importFrom,
            });
        }
    });
};
exports.renderImports = renderImports;
const hasPathParams = (op) => {
    return !!op.parameters?.find(e => {
        if ('in' in e) {
            return e.in === 'path';
        }
        return false;
    });
};
exports.hasPathParams = hasPathParams;
const hasFileUpload = (op) => {
    const reqBody = op.requestBody;
    if (!reqBody)
        return false;
    const contentKey = 'multipart/form-data';
    if (!(contentKey in reqBody.content))
        return false;
    return true;
};
exports.hasFileUpload = hasFileUpload;
const extractOperationTypeNames = (operation) => {
    const isMutation = (0, exports.isOperationMutation)(operation['x-method'], operation);
    const response = (0, case_1.pascal)(operation.operationId + '_response');
    const responseData = (0, case_1.pascal)(operation.operationId + '_response_data');
    const responseMeta = (0, case_1.pascal)(operation.operationId + '_response_meta');
    let request = (0, case_1.pascal)(operation.operationId + '_request');
    if (isMutation) {
        return { request, response, responseData, responseMeta };
    }
    if (!operation.requestBody) {
        request = null;
    }
    return {
        request,
        response,
        responseData,
        responseMeta,
    };
};
exports.extractOperationTypeNames = extractOperationTypeNames;
